#include <assert.h>
#include "Graph.h"

NodePtrDLList    Graph::leafNodeList ;
NodePtrDLList    Graph::intermediateNodeList ;
StringNodePtrDLListPtrAVLMap    Graph::queryCaching(0) ;

NodePtrDLList*
Graph::Search( Attribute& anAttribute )
{
	if ( Graph::leafNodeList.empty() && Graph::intermediateNodeList.empty() )
		return NULL ;

	NodePtrDLList* newList = new NodePtrDLList ;

	assert( newList ) ;

	if ( !Graph::leafNodeList.empty() )
	{
		// First search the leaf List
		for ( Pix i = Graph::leafNodeList.first() ; i ; 
			 Graph::leafNodeList.next(i) )
		{
			NodePtr tmp = Graph::leafNodeList(i) ;

			if ( tmp->SearchAttribute( anAttribute ) )
				newList->append( tmp ) ;
		}
	}

	if ( !Graph::intermediateNodeList.empty() )
	{
		// Next search the intermediate node List
		for ( Pix i = Graph::intermediateNodeList.first() ; i ; 
			 Graph::intermediateNodeList.next(i) )
		{
			NodePtr tmp = Graph::intermediateNodeList(i) ;

			if ( tmp->SearchAttribute(anAttribute) )
				newList->append( tmp ) ;
		}
	}

	if ( newList->empty() )
	{
		delete newList ; return NULL ;
	}

	return newList ;
}

NodePtrDLList*
Graph::Search( NodePtrDLList& aNodeList, Attribute& anAttribute )
{
	if ( aNodeList.empty() ) 
		return NULL ;

	NodePtrDLList* newList = new NodePtrDLList ;

	assert( newList ) ;

	if ( !aNodeList.empty() )
	{
		// First search the leaf List
		for ( Pix i = aNodeList.first() ; i ; aNodeList.next(i) )
		{
			NodePtr tmp = aNodeList(i) ;

			if ( tmp->SearchAttribute( anAttribute ) )
				newList->append( tmp ) ;
		}
	}

	if ( newList->empty() )
	{
		delete newList ; return NULL ;
	}

	return newList ;
}	

NodePtr 
Graph::SearchForNode( String& str )
{
        NodePtr tmp ;

        // Search the leaf list first
        if ( tmp = Graph::leafNodeList.retrieve(str) )
        {
                return tmp ;
        }

        // Else search the intermediate node list
        if ( tmp = Graph::intermediateNodeList.retrieve(str) )
        {
                return tmp ;
        }

        return NULL ;
}

void
Graph::ConstructGraph()
{
	// cout << "Invoking Graph " << endl ;

	for ( Pix i = Graph::intermediateNodeList.first() ; i ;
			Graph::intermediateNodeList.next(i) )
	{
		NodePtr node = Graph::intermediateNodeList(i) ;

		// Traverse the parent list
		for ( Pix j = node->parentList.first() ; j ; 
			node->parentList.next(j) )
		{
			EdgePtr edge = node->parentList(j) ;

			edge->nodePtr = Graph::SearchForNode(edge->nodeName) ;
		}	

		// Traverse the child list
		for ( j = node->childList.first() ; j ; 
			node->childList.next(j) )
		{
			EdgePtr edge = node->childList(j) ;

			edge->nodePtr = Graph::SearchForNode(edge->nodeName) ;
		}	
	}
}
