#include <assert.h>
#include "QueryTree.h"
#include "Graph.h"

NodePtrDLList*
QueryTree::traverseEvaluate( int nodeNum, NodePtrDLList* input)
{
	if ( !nodeNum ) return NULL ;

	TreeNodePtr node = tree(tree.elem(nodeNum)) ;

	assert ( node ) ; 
	if ( node->isLeaf() )
	{
		NodePtrDLList* ptr = NULL ;

		if ( input )
		{
			ptr = Graph::Search(*input,node->info.attrName,
						node->info.attrValue) ;
		}
		else
		{
			ptr = Graph::Search(node->info.attrName,
						node->info.attrValue) ;
		}

		return ptr ; 
	}

	NodePtrDLList* list1 = NULL ;
	NodePtrDLList* list2 = NULL ; 

	if ( node->oper == OR ) 
	{
		list1 = traverseEvaluate(node->left, NULL ) ;
		list2 = traverseEvaluate(node->right,NULL ) ;

		if ( list1 && list2 )
		{
			list1->join_unique(*list2) ;

			list2->clear() ; delete list2 ; 
		}	
		else if ( list2 )
			list1 = list2 ; 
			
	}
	else if ( node->oper == AND )
	{
		list2 = traverseEvaluate(node->left,input) ; 

		if ( list2 ) 
		{
			list1 = traverseEvaluate(node->right,list2) ; 
				
			list2->clear() ; delete list2 ; 
		}	
	}

	return list1 ; 
}

NodePtrDLList*
QueryTree::evaluateSubQuery(int nodeNum, NodePtr& queryNode)
{
	if ( !nodeNum )  return NULL ;
	
	QueryTree* subQuery = new QueryTree ;
	assert( subQuery ) ;
	subQuery->head = nodeNum ;  
	subQuery->tree = tree ;  

	NodePtrDLList* queryList = subQuery->findCandidateQueries() ;


	if ( queryList )
	{

#ifdef DEBUG
		cout << "----------Candidate Queries----------" << endl ;
		queryList->print() ;
		cout << "----------Candidate Queries----------" << endl ;
#endif

		NodePtrDLList*	answer = subQuery->findAnswerQuery(*queryList,queryNode) ; 

		if ( answer )
                {
                        // cout << "Found Cached Sub " ;
			// subQuery->printQuery() ;
                }

		queryList->clear() ; delete queryList ; 

		if ( answer ) 
		{
			delete subQuery ; 
			return answer ;
		}
	}

	delete subQuery ; 

	queryNode = NULL ; return NULL ; 
}	

NodePtrDLList*
QueryTree::evaluateQueryFromExistQuery()
{
	NodePtr node ;

	int saveFlag ; 

	// Find whether cached query is there or not  
	NodePtrDLList* list = evaluateSubQuery(head,node) ;  
	
	if ( !list || list->isQueryList() )	// if not evaluate in terms of other sub queries
	{
		if ( list )  delete list ; 
		list = traverseEvaluateSubQuery(head,NULL,node,saveFlag) ;
	}

	return list ; 
}


NodePtrDLList*
QueryTree::traverseEvaluateSubQuery( int nodeNum, NodePtrDLList* input, NodePtr& queryNode,
					int& saveFlag)
{
	if ( !nodeNum ) 
	{
		queryNode = NULL ; 
		return NULL ;
	}

	TreeNodePtr node = tree(tree.elem(nodeNum)) ;
	assert ( node ) ; 

	NodePtrDLList*	list1 = NULL ;
	NodePtrDLList*	list2 = NULL ; 
	NodePtr	      	leftQueryNode = NULL ;			
	NodePtr	      	rightQueryNode = NULL ;			
	double		leftSelectivity = 0 ;
	double		rightSelectivity = 0 ; 
	int		leftSaveFlag = 0 ;
	int		rightSaveFlag = 0 ; 

	if ( node->oper == OR ) 
	{
		// Find whether cached left sub query is there or not  
		list1 = evaluateSubQuery(node->left,leftQueryNode) ;  

		// Actually at this point we might know which sub queries
		// have to applied (computed actually) if the answer returned 
		// is a another set of queries in the case the current 'node'
		// is a apply node. We don't have to go thro' the extra
		// search of index for finding the left and right sub queries 

		// The input is NULL because of the fact that the 
		// a sub query of AND that is evaluated for the first time
		// should be saved. Subsequent times we don't have to 
		// evaluate.

		if ( !list1 || list1->isQueryList() )
		{
			// if not, check cached sub queries of left sub query 
			if ( list1 ) { list1->clear() ; delete list1 ; }  

			list1 = traverseEvaluateSubQuery( node->left, NULL, leftQueryNode,
							leftSaveFlag ) ; 
		}

		if ( list1 )	leftSelectivity = list1->length() ; 

		// Find whether cached right sub query is there or not  
		list2 = evaluateSubQuery(node->right,rightQueryNode) ;  

		if ( !list2 || list2->isQueryList() )	
		{
			if ( list2 ) { list2->clear() ; delete list2 ; }   

			list2 = traverseEvaluateSubQuery( node->right, NULL, rightQueryNode,
					rightSaveFlag) ; 	
		}

		if ( list2 )	rightSelectivity = list2->length() ; 

		if ( list1 && list2 )
		{
			// Eliminate Duplicates
			list1->join_unique(*list2) ;

			list2->clear() ; delete list2 ; 
		}	
		else if ( list2 )
			list1 = list2 ; 
	}
	else if ( node->oper == AND )
	{
		// Find whether cached left sub query is there or not  
		list1 = evaluateSubQuery(node->left,leftQueryNode) ;  

		if ( !list1 || list1->isQueryList() )
		{
			// if not, check cached sub queries of left sub query 
			if ( list1 ) { list1->clear() ; delete list1 ; }  

			list1 = traverseEvaluateSubQuery( node->left, NULL, leftQueryNode,
							leftSaveFlag ) ; 
		}

		if ( list1 )	leftSelectivity = list1->length() ; 

		// Find whether cached right sub query is there or not  
		list2 = evaluateSubQuery(node->right,rightQueryNode) ;  

		if ( !list2 || list2->isQueryList() )	
		{
			// if not, check cached sub queries of left sub query 
			if ( list2 ) { list2->clear() ; delete list2 ; }   

			list2 = traverseEvaluateSubQuery( node->right, NULL, rightQueryNode,
					rightSaveFlag) ; 	
		}

		if ( list2 )	rightSelectivity = list2->length() ; 

		if ( list1 && list2 )
		{
			list1->join_intersect(*list2) ;

			list2->clear() ; delete list2 ; 
		}	
		else if ( list1 )
		{
			list1->clear() ; delete list1 ; list1= NULL ;
		}
		else
		{
			list2->clear() ; delete list2 ;  list2 = NULL ; 
		}
	}
	else if ( node->isLeaf() )
	{
		list1 = evaluateSubQuery(nodeNum, leftQueryNode) ; 	

		if ( !list1 )
		{
			//cout << "Doing a brute force evaluation " << endl ; 
			list1 = traverseEvaluate(nodeNum,input) ; 
		}
		else
			saveFlag = 1 ; 

		if ( list1 )	leftSelectivity = list1->length() ; 
	}

	double leftPerCent = double(leftSelectivity*100)/Graph::NumberOfNodes() ; 
	double rightPerCent = double(rightSelectivity*100)/Graph::NumberOfNodes() ; 

	// if ( node->isLeaf() && !saveFlag && leftPerCent < 25 )
	if ( node->isLeaf() )
	{
		queryNode = cacheQuery(list1,nodeNum) ;

		Graph::intermediateNodeList.append( queryNode ) ; 	
	}
	else 
	{
		// Create a new query Node 	
		queryNode = new Node ;
		
		queryNode->setNoLeaf() ; queryNode->setApply() ; 

		queryNode->queryTree = new QueryTree ; assert( queryNode->queryTree ) ; 

		queryNode->queryTree->tree = this->tree ;

		queryNode->queryTree->head = nodeNum ;

		queryNode->oper = node->oper ; 

		// Create a new name 
		queryNode->name = "_" ;
		queryNode->name += leftQueryNode->name ;
		queryNode->name += rightQueryNode->name ;

		// Insert the left child 
		EdgePtr leftEdge = new Edge ; assert( leftEdge ) ; 

		leftEdge->nodePtr = leftQueryNode ;  

		queryNode->childList.append( leftEdge ) ;

		// Insert the right child 
		EdgePtr rightEdge = new Edge ; assert( rightEdge ) ; 

		rightEdge->nodePtr = rightQueryNode ;  

		queryNode->childList.append( rightEdge ) ;

		// Insert this node into the parent list
		EdgePtr anEdge = new Edge ; assert( anEdge ) ;

		leftQueryNode->parentList.append( anEdge ) ;

		rightQueryNode->parentList.append( anEdge ) ; 

		// Update the hashing function
		traverseCache( nodeNum, queryNode ) ; 

		Graph::intermediateNodeList.append( queryNode ) ; 	
	}

	return list1 ; 
}

void
QueryTree::traversePrint( int nodeNum )
{
	if ( !nodeNum ) return ; 

	TreeNodePtr node = tree(tree.elem(nodeNum)) ;

	assert ( node ) ;

	if ( node->isLeaf() )
	{
		cout << '(' 
		     << node->info.attrName 
		     << ',' 
		     << node->info.attrValue 
		     << ") "  ; 

		return ;
	}


	if ( node->oper == OR )
		cout << "OR " ;
	else if ( node->oper == AND )
		cout << "AND " ;

	traversePrint( node->left ) ;
	traversePrint( node->right ) ; 
}

int
QueryTree::traverseEqual( int treeNum1, int treeNum2, QueryTree& aQueryTree)
{
	if ( !treeNum1 && !treeNum2 ) return 1 ;

	if ( (treeNum1 && !treeNum2) || (!treeNum1 && treeNum2) ) return 0 ; 

	TreeNodePtr node1 = tree(tree.elem(treeNum1)) ;
	TreeNodePtr node2 = aQueryTree.tree(aQueryTree.tree.elem(treeNum2)) ;

	// Current node is equal
	if ( *node1 == *node2 )
	{
		// Check if subtrees are equal 
		if ( traverseEqual(node1->left,node2->left, aQueryTree) )
		{
			return traverseEqual ( node1->right, node2->right, aQueryTree ) ;
		}
		else if ( traverseEqual(node1->left,node2->right, aQueryTree) )
		{
			return traverseEqual ( node1->right, node2->left, aQueryTree ) ;
		}

		return 0 ; 
	}
	else
		return 0 ; 

	return 1 ; 
}

NodePtr
QueryTree::cacheQuery(NodePtrDLList* answer, int nodeNum)
{
	// Create a query node with appropriate details
	NodePtr queryNode = new Node ;

	queryNode->queryTree = new QueryTree ; assert( queryNode->queryTree ) ; 
	
	queryNode->queryTree->tree = this->tree ;
	
	queryNode->queryTree->head = nodeNum ; 

	queryNode->setNoLeaf() ;

	queryNode->setNoApply() ; 

	queryNode->oper = NOP ;

	if ( answer )
	{
		queryNode->name = "_" ;

		// Fill the edge list of the query node so that
		// the answer need not be computed. 
		for ( Pix i = answer->first() ; i ; answer->next(i) )
		{
			NodePtr tmp = (*answer)(i) ; 

			EdgePtr newEdge = new Edge ;

			newEdge->nodeName = tmp->name ;

			newEdge->nodePtr = Graph::SearchForNode(tmp->name) ; 

			queryNode->childList.append(newEdge) ; 

			queryNode->name += tmp->name ; 
		}
	}
	else
	{
	       queryNode->name = "junk" ;
	}

	traverseCache(nodeNum,queryNode) ;

	return queryNode; 
}

void
QueryTree::traverseCache( int nodeNum, NodePtr queryNode )
{
	if ( !nodeNum ) return ;

	TreeNodePtr node = tree(tree.elem(nodeNum)) ;

	assert ( node ) ; 

	if ( node->isLeaf() )
	{
		NodePtrDLListPtr tmp = Graph::queryCaching[node->info.attrName] ; 

		if ( !tmp )
		{
			tmp = new NodePtrDLList ;
		}
		
		tmp->append( queryNode ) ;

		Graph::queryCaching[node->info.attrName] = tmp ; 

#ifdef DEBUG
		cout << "---------------------------" << endl ;
		tmp->print() ; 
		cout << "---------------------------" << endl ;
#endif
	}

	// Traverse the left subtree
	traverseCache(node->left, queryNode) ; 

	traverseCache(node->right, queryNode) ; 

	return ; 
}
	
NodePtrDLList*
QueryTree::traverseQuerySearch( int nodeNum, NodePtrDLList* input)
{
	if ( !nodeNum ) return NULL ;

	TreeNodePtr node = tree(tree.elem(nodeNum)) ;

	assert ( node ) ; 
	if ( node->isLeaf() )
	{
		// If there is no intermediate queries 
		if ( Graph::queryCaching.empty() )
			return NULL ; 

		// Search the query caching for this attribute
		NodePtrDLList* tmpList = Graph::queryCaching[node->info.attrName] ;

		if ( tmpList && !tmpList->empty() ) 
		{	
			NodePtrDLList* ansList = new NodePtrDLList ; assert ( ansList ) ; 
			*ansList = *tmpList ; 
			return ansList ; 
		}

		return NULL ; 
	}

	NodePtrDLList* list1 = NULL ;
	NodePtrDLList* list2 = NULL ; 

	if ( node->oper == OR ) 
	{
		list1 = traverseQuerySearch(node->left, NULL ) ;
		list2 = traverseQuerySearch(node->right,NULL ) ;

		if ( list1 && list2 )
		{
			list1->join_unique(*list2) ;

			list2->clear() ; delete list2 ; 
		}	
		else if ( list2 )
			list1 = list2 ; 
	}
	else if ( node->oper == AND )
	{
		list2 = traverseQuerySearch(node->left,input) ; 

		if ( list2 ) 
		{
			list1 = traverseQuerySearch(node->right,list2) ; 

			list1->join_intersect(*list2) ;

			list2->clear() ; delete list2 ; 
		}	
	}

	return list1 ; 
}

NodePtrDLList* 
QueryTree::findAnswerQuery( NodePtrDLList& candidateList, NodePtr& queryNode )
{
	for ( Pix i = candidateList.first() ; i ; candidateList.next(i) )
	{
		NodePtr tmp = candidateList(i) ;  

		if ( *this == *(tmp->queryTree) )
		{
			NodePtrDLList* newList = new NodePtrDLList ;

			for ( Pix j = tmp->childList.first() ; j ; 
					tmp->childList.next(j) )
			{
				EdgePtr edge = tmp->childList(j) ;  
					
				newList->append(edge->nodePtr) ;

				if ( !edge->nodePtr->isLeaf() )
					newList->setQueryList() ;	
			}

			queryNode = tmp ;  

			return newList ;  
		}
	}

	return NULL ;
}
