#ifndef __QUERYTREE_H
#define	__QUERYTREE_H

#include <String.h>
#include "TreeNodePtr.SLList.h"
#include "NodePtr.DLList.h"

class QueryTree
{
	public :

		QueryTree() : head(0) { } ;

		int
		operator == ( QueryTree& aQueryTree ) 
		{
			return traverseEqual(head,aQueryTree.head,aQueryTree) ;
		}

		// Brute force evaluation
		NodePtrDLList*
		evaluateQuery() 
		{
			return traverseEvaluate(head,NULL) ;
		}

		// Evaluate a query in terms of existing queries
		NodePtrDLList*
		evaluateQueryFromExistQuery() ;  


		// Find candidate queries
		NodePtrDLList*
		findCandidateQueries()
		{
			return traverseQuerySearch(head,NULL) ; 
		}

		// Find the node that contains the required
		// answer
		NodePtrDLList*
		findAnswerQuery(NodePtrDLList& candidateList,NodePtr& queryNode) ; 

		// Cache the query into the existing graph
		NodePtr
		cacheQuery(NodePtrDLList* answer, int nodeNum) ;  

		void
		printQuery() 
		{
			cout << "Query: "  ;

			traversePrint(head) ; 
			cout << endl ; 
		}	

		int			head ; 	

		TreeNodePtrSLList	tree ; 
		
	private :

		NodePtrDLList*	
		traverseEvaluate( int nodeNum, NodePtrDLList* input ) ;

		NodePtrDLList*	
		traverseEvaluateSubQuery( int nodeNum, NodePtrDLList* input, NodePtr& node,
						int& saveFlag ) ;

		NodePtrDLList*	
		traverseQuerySearch( int nodeNum, NodePtrDLList* input ) ;

		void
		traversePrint( int nodeNum ) ;

		int
		traverseEqual( int nodeNum1, int nodeNum2, QueryTree& aQueryTree) ; 

		void
		traverseCache( int nodeNum, NodePtr queryNode) ; 

		NodePtrDLList*
		evaluateSubQuery(int nodeNum, NodePtr& queryNode) ;  
} ;


#endif
