%{
/* *****************************************************************************
 * Lexical analyser for the node descriptions.
 */

#include <stdio.h>
#include <stdlib.h>

#include "y.tab.h"
#include "global.h"
#include "error.h"
#include <search.h>		/* for binary search */

extern charPtr yylval ;
%}

delim   [\t\n ]
white   {delim}+
letter  [_A-Za-z]
digit   [0-9]
number  [0-9]({digit})*
id      {letter}({letter}|{digit})*

%%
"/*"			{
                           int i, lasti = 0;

			   while ((i = input() ) && i != EOF)
			   {
			      
			      if (i == '/' && lasti == '*')
				 break;		/* recognize comment	*/
			      lasti = i;
			   }

			   if (i == 0 || i == EOF)
			      yyerror("End of file in comment\n");
			}

\"(\\.|[^\"])*\"	return STRING;


\"(\\.|[^\"])*[\n]	{
   			   yyerror("Adding missing \" to string constant\n");
			   return STRING;
			}


{number}		{ yylval = yytext ; return ICON; }

"{"			return LC;
"}"			return RC;
":"			return COLON;
";"			return SEMI;
","			return COMMA;
"="			return EQUAL;
"("			return LB ;
")"			return RB ;
"+"			return EOR ;	
"*"			return EAND ; 

{id}			return id_or_keyword(yytext);

{white}			; /* Ignore white space */

.			yyerror("Illegal character <%s>\n", yytext);

%%

/****************************************************************/

typedef struct
{
   char *name;
   int   val;
}
KWORD;

KWORD Ktab[] =			/* Alphabetic keywords		*/
{
   { "I",	  OP		},
   { "U",	  OP		},
   { "and",	  OP		},
   { "attribute", ATTRIBUTE	},
   { "children",  CHILDREN	},
   { "end",	  END		},
   { "intermediate", INTER	},
   { "leaf",	  LEAF		},
   { "operator",  OPERATOR	},
   { "or",	  OP		},
   { "parents",   PARENTS	},
   { "start",	  START		}
};


int cmp(KWORD *a, KWORD *b)
{
   return strcmp(a->name, b->name);
}

int id_or_keyword(char *str)
{
   KWORD *p;
   KWORD  dummy;

   yylval = yytext ;
   dummy.name = str;


   p = bsearch(&dummy, Ktab, sizeof(Ktab)/sizeof(KWORD), sizeof(KWORD), cmp);
   return (p ? p->val : ID);
}

int yywrap() {return 1; }
