#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <stdlib.h>
#include <String.h>
#include "y.tab.h"
#include "NodePtr.DLList.h"
#include "TreeNodePtr.SLList.h"
#include "Graph.h"
#include "QueryTree.h"

extern "C" {

	char*	initValue( char * ) ;
	char*	appendValue( char * ) ;
	char*	initName( char * ) ;
	char*	initNodeName( char * ) ;
	char*	initStartTime( char * ) ;
	char*	initEndTime( char * ) ;
	char*	initOperator( char * ) ; 
	char*	initParent( char * ) ;
	char*	initChildren( char * ) ;
	char*	appendParent( char * ) ;
	char*	appendChildren( char * ) ;
	char*	newAttribute( char*, char *) ;  
	char* 	newLeafNode( char* ) ; 
	char* 	newIntermediateNode( char* ) ; 

	char*	queryLeafNode(char*,char*) ; 	
	char*	queryIntermediateNode(char*,char*,int) ; 
	char*	querySetHead(char*) ; 
	void	ConstructGraph() ; 
	void	processQuery() ; 
} ; 

static const char queryOutputFile[] = "/tmp/_query_result" ;
static const char browserPath[] = "/u/r/a/ramasamy/public/784/code/c/xanim" ;  
static const char browserArg0[] = "/u/r/a/ramasamy/public/784/code/c/xanim" ;
static const char browserArg1[] = "/u/s/a/samit/public/commercial.qt" ;

static String			attributeValue ;
static String			attributeName ;
static String			nodeName ;	
static String			startTime ;
static String			endTime	;
static String			operStr ;	
static OperatorType		oper ;	

static AttributePtrSLList	attrList ;
static EdgePtrDLList		parList ;
static EdgePtrDLList		childList ; 

QueryTree*			userQuery = NULL ;		

char*
initValue( char* str )
{
	attributeValue = str ;

	attributeValue.downcase() ;

	return (char *)attributeValue.chars() ;
}

char*
appendValue( char* str )
{
	attributeValue += ' ' ;
	
	attributeValue += str ;

	attributeValue.downcase() ;

	return (char *)attributeValue.chars() ;
}

char*
initName( char* str )
{
	attributeName = str ;

	attributeName.downcase() ;

	return (char *)attributeName.chars() ;
}

char*
initNodeName( char* str )
{
	nodeName = str ;

	nodeName.downcase() ;

	return (char *)nodeName.chars() ;
}

char*
initStartTime( char* str )
{
	startTime = str ;

	return (char *)startTime.chars() ;
}

char*
initEndTime( char* str )
{
	endTime = str ;

	return (char *)endTime.chars() ;
}

char*
initOperator( char* str )
{
	if ( !strcmp(str,"I") )
		oper = INTERSECTION ;
	else if ( !strcmp(str,"U") )
		oper = UNION ;
	else if ( !strcmp(str,"or") ) 
		oper = OR ;
	else if ( !strcmp(str,"and") )
		oper = AND ;
	else
		oper = NOP ;

	operStr = str ;

	// cout << operStr << endl ;

	return (char *)operStr.chars() ;
}

char*
initParent( char* str )
{
	parList.clear() ;

	EdgePtr	edge = new Edge ;
	
	assert( edge ) ;

	edge->nodeName = str ;

	parList.append(edge) ;

	return (char *)edge->nodeName.chars() ; 
}

char*
appendParent( char* str )
{
	EdgePtr	edge = new Edge ;
	
	assert( edge ) ;

	edge->nodeName = str ;

	parList.append(edge) ;

	return (char *)edge->nodeName.chars() ; 
}

char*
initChildren( char* str )
{
	childList.clear() ; 

	EdgePtr	edge = new Edge ;
	
	assert( edge ) ;

	edge->nodeName = str ;

	childList.append(edge) ;

	return (char *)edge->nodeName.chars() ; 
}

char*
appendChildren( char* str )
{
	EdgePtr	edge = new Edge ;
	
	edge->nodeName = str ;

	assert( edge ) ;

	childList.append(edge) ;

	return (char *)edge->nodeName.chars() ; 
}

char *
newAttribute( char* attrName, char* attrValue )
{
	Attribute*	attribute = new Attribute(attrName,attrValue) ; 

	assert( attribute ) ; 
	
	// cout << *attribute << endl ;

	attrList.append( attribute ) ; 

	return attrName ;
}

char*
newLeafNode( char* name )
{
	Node*	node = new Node ;

	assert( node ) ;

	node->name = name ;
	
	node->startTime = atoi(startTime.chars()) ;

	node->endTime = atoi(endTime.chars()) ;

	node->attrList.join( attrList ) ;

	node->setLeaf() ;

	// cout << *node << endl ;

	Graph::leafNodeList.append(node) ;

	attrList.clear() ; childList.clear() ; parList.clear() ; 

	return (char *)node->name.chars() ;
}

char*
newIntermediateNode( char* name )
{
	Node*	node = new Node ;

	assert( node ) ;

	node->name = name ;

	node->oper = oper ;

	node->childList.join( childList ) ;

	node->parentList.join( parList ) ; 

	node->attrList.join( attrList ) ; 

	node->setNoLeaf() ;

	// cout << *node << endl ;

	Graph::intermediateNodeList.append(node) ;

	attrList.clear() ; childList.clear() ; parList.clear() ; 

	return (char *)node->name.chars() ;
}

char*
queryLeafNode( char * attrName, char* attrValue )
{
	if ( !userQuery )
		userQuery = new QueryTree ;

	// Create a new tree Node
	TreeNodePtr node = new TreeNode ;

	assert( node ) ;

	node->info.attrName = attrName ;
	node->info.attrValue = attrValue ; 

	node->setLeaf() ; 

	userQuery->tree.append( node ) ;

	int nodeNum = userQuery->tree.length() ;

	strstream str ;

	str << nodeNum << '\0' ; 	
	
	node->nodeNum = str.str() ; 

	return (char *)(node->nodeNum.chars()) ; 
}

char*
querySetHead(char *headNum)
{
	userQuery->head = atoi(headNum) ;

	operStr = headNum ;

	// userQuery->printQuery() ; 

	return (char*)operStr.chars() ; 
}

void
processQuery()
{
	NodePtrDLList* answer = userQuery->evaluateQueryFromExistQuery() ;  	
		
#if 0
	if ( queryList )
	{
		NodePtrDLList* answer = userQuery->findAnswerQuery(*queryList) ; 
		
		if ( answer )
		{
			// cout << "Found Cached Query " << endl ; 
			// cout << "Answer : "  << endl ; 
			// answer->print() ; 
		}

		queryList->clear() ; delete queryList ; 
		
		if ( answer ) 
		{
			answer->clear() ; delete answer ;  
			return ; 
		}
	}

	// cout << "Doing a Brute Force Evaluation " << endl ;

	// Brute Force Evaluation
	NodePtrDLList* bottomUp = userQuery->evaluateQuery() ;

	// cout << "Answer : " << endl ; 

	if ( bottomUp )
	{
		// bottomUp->print() ;
	}

	userQuery->cacheQuery(bottomUp) ;	// Saves the query  

	if ( bottomUp )
	{
		bottomUp->clear() ; delete bottomUp ; 	
	}
#endif

	if ( answer && !answer->empty() )
	{
		ofstream fout(queryOutputFile) ;
		// cout << "Answer: "  << endl ; 
		// answer->print() ; 

		fout << *answer ; 

		/*
		cout << "Starting the browser..." << endl ; 

		int pid = -1 ; 

		if ( (pid = vfork()) < 0 )
		{
			cerr << "Unable to fork a process." << endl ;
		}
		else if ( pid == 0 )
		{
			execlp(browserPath,browserArg0,browserArg1,(char*)0) ; 	
		}
		*/
	}
	else 
	{
		ofstream fout(queryOutputFile) ;
		fout.close() ;
	}	
		

	delete answer ; 

	cout << "Graph: " << endl ; 
	Graph::leafNodeList.print() ;
	Graph::intermediateNodeList.print() ; 

	userQuery = NULL ; 
}

char*
queryIntermediateNode(char* leftChild, char* rightChild, int flag)
{
	// Create a new tree Node
	TreeNodePtr node = new TreeNode ;

	assert( node ) ;

	node->setNoLeaf() ; 

	node->oper = flag ? OR : AND ; 

	node->left = atoi(leftChild) ;
	node->right = atoi(rightChild) ; 

	userQuery->tree.append( node ) ;

	int nodeNum = userQuery->tree.length() ;

	strstream str ;

	str << nodeNum << '\0' ; 	
	
	node->nodeNum = str.str() ; 

	return (char *)(node->nodeNum.chars()) ; 
}

void
ConstructGraph()
{	
	Graph::ConstructGraph() ;
}
