%{
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#include "y.tab.h"
#include "global.h"
#include "error.h"

extern char *yytext;			/* generated by lex	*/
extern int   yylineno;			/* generated by lex	*/

extern	char* initValue(char*) ;
extern	char* appendValue(char*) ;
extern	char* initName(char*) ;
extern	char* initNodeName(char*) ;	
extern	char* initStartTime( char * ) ;
extern	char* initEndTime( char * ) ;
extern	char* initOperator( char *) ; 	
extern	char* initParent( char * ) ;	
extern	char* initChildren( char * ) ;	
extern	char* appendParent( char * ) ;	
extern	char* appendChildren( char * ) ;	
extern	char* newAttribute( char*, char *) ;
extern	char* newLeafNode( char* ) ; 
extern	char* newIntermediateNode( char* ) ; 

extern	char* queryLeafNode(char*,char*) ;  
extern	char* queryIntermediateNode(char*,char*,int) ;  
extern	char* querySetHead(char*) ; 	
extern	void  processQuery() ;
extern	void  ConstructGraph() ; 	

#ifdef  YYSTYPE
#undef  YYSTYPE
#endif

#define YYSTYPE	     charPtr    	/* Value stack type	*/
%}

%token ATTRIBUTE
%token CHILDREN
%token END
%token INTER
%token LEAF
%token OPERATOR
%token PARENTS
%token START

%token STRING
%token ICON
%token LC
%token RC
%token SEMI
%token COLON
%token COMMA
%token OP
%token ID
%token EQUAL
%token LB
%token RB
%token EAND
%token EOR


%start program

%%

program	:	load { ConstructGraph() ;} queries 
	;

load 	:	leaf					
	|	intermediate
	|	load leaf
	|	load intermediate
	;

leaf	:	LEAF LC attrs start end RC node_id SEMI {$$ = newLeafNode($7);}  
	;

intermediate :	INTER LC attrs operator parChild RC node_id SEMI { $$ = newIntermediateNode($7);}; 
	;

attrs	:	attr
	|	attrs attr
	;

attr	:	ATTRIBUTE attr_id COLON ids SEMI { $$ = newAttribute($2,$4) ; } 
	;

ids	:	ID	{ $$ = initValue($1) ; } 
	|	ids ID	{ $$ = appendValue($2) ; } 
	;

attr_id	: 	ID	{ $$ = initName($1) ; }  
	;

node_id	: 	ID	{ $$ = initNodeName($1) ; }  
	;

start	:	START COLON stime SEMI { $$ = $3 ; }
	;

end	:	END COLON etime SEMI { $$ = $3 ; } 
	;

stime	:	ICON	{ $$ = initStartTime($1) ; } 
	;

etime	:	ICON	{ $$ = initEndTime($1) ; } 
	;

parChild :	parents children
	|	children parents
	|	parents
	|	children
	;

parents	:	PARENTS COLON pnames SEMI
	;

children:	CHILDREN COLON cnames SEMI
	;

pnames	:	ID	{ $$ = initParent($1) ; } 
	|	pnames COMMA ID { $$ = appendParent($3) ; }
	;

cnames	:	ID	{ $$ = initChildren($1) ; } 
	|	cnames COMMA ID { $$ = appendChildren($3) ; }
	;

operator:	OPERATOR COLON oper SEMI
	;

oper	:	OP { $$ = initOperator($1) ; }  
	;

queries :	query 
	|	queries query
	;

query	:	expr SEMI 		{ $$ = querySetHead($1); processQuery(); } 
	;

expr	:	expr EOR term		{ $$ = queryIntermediateNode($1,$3,1) ;}
	|	term			{ $$ = $1 ;}
	;

term	:	term EAND factor	{ $$ = 	queryIntermediateNode($1,$3,0) ;}
	|	factor			{ $$ = $1 ;}
	;

factor	:	LB expr RB		{ $$ = $2 ;} 
	|	qspec			{ $$ = $1 ;} 
	;

qspec	:	exprid EQUAL exprids	{ $$ = queryLeafNode($1,$3) ;} 
	;

exprids	:	ID			{ $$ = initValue($1) ;} 	 	
	|	exprids ID		{ $$ = appendValue($2) ;} 	
	;

exprid	:	ID			{ $$ = initName($1) ;} 
	;
%%

#if 0
   leaf {
        attribute <attrib-name> : <value>  ; ...

        /* For video play times */
        start : <value> ;
        end :   <value> ;

} A ; /* Name required for each node to refer them as children
         and parents. */

intermediate {

attribute actor "value" ;
attribute salary "10" ;

operator: or

parents: D, E ; 
children: A, B ; 

} C ;



#endif
